from .EEGNet import EEGNet
from .EEGConformer import EEGConformer
from .FBCNet import FBCNet
from .IFNet import IFNet, IFNetAdamW
from .LightConvNet import LightConvNet
from .MSVTNet import MSVTNet, JointCrossEntoryLoss
from .ShallowConvNet import ShallowConvNet
from .DeepConvNet import DeepConvNet
